#!/usr/bin/perl -w
#
# Copyright (C) 2006  Pascal Giard
#
# Author:             Pascal Giard <evilynux@gmail.com>
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

use strict;

use Getopt::Long;
# Intl
use POSIX;
use Locale::TextDomain ('joellisme');
########### EXPERIMENTAL ###################
use Locale::gettext;
textdomain('joellisme');
eval { require c::stuff };
if ($@) {
    bind_textdomain_codeset('joellisme', "UTF-8");
} else {
    c::stuff::bind_textdomain_codeset('joellisme', "UTF-8");
}
binmode(STDOUT, ":utf8");
binmode(STDERR, ":utf8");
#######################################

#use FindBin ();
#use lib "$FindBin::Bin/joellisme";        # joellisme isn't installed
#use lib "$FindBin::Bin/../lib/joellisme"; # joellisme is installed
#use config;

use constant FALSE => 0;
use constant TRUE => !FALSE;

# Global variables
my $NAME = "joellisme";
my $VERSION = "0.6.0";
use vars qw(%opt $glade_dir $xml_root $text);

# Prototypes
sub gui();
sub btnQuitClicked();
sub parseText($);
sub msgDebug($);
sub msgError($);
sub process($);
sub test($$);
sub selfTest();
sub showSuggestion(%);
sub help();
sub version();
sub main_cmdline();

GetOptions( "debug"   => \$opt{'debug'},
	    "nogui"   => \$opt{'nogui'},
	    "help"    => \$opt{'help'},
	    "version" => \$opt{'version'},
	    "<>" => \&parseText
	  );

help() if( defined($opt{'help'}) );
version() if( defined($opt{'version'}) );

selfTest();

# The graphical interface is to be shown
if (!defined $opt{'nogui'}) {
  # Trying to load the GladeXML perl module
  eval "use Gtk2::GladeXML;";
  if ($@) {
    $opt{'nogui'} = TRUE;
    msgError( __"The Gtk2::GladeXML perl module needs to be installed!\n".
              "Please refer to README file.\n".
              "-> Falling back to command line mode.\n");
  } else {
      # Useful for constants
      use Gtk2::Pango;
  }
}

# FIXME, use proper perl module instead of doing those crazy things
my $current_path = $0;
$current_path =~ s/joellisme(\.pl)?$//;

if ($0 =~ /^\.\//) {
  $glade_dir= '.';
} else {
  $glade_dir = "$current_path../share/joellisme/glade";
  $glade_dir= '.' unless( -f $glade_dir );
}

# Let's begin
if (defined $opt{'nogui'}) {
  unless ( defined($text) ) {
    print STDERR "\n";
    print STDERR __"*** You need to specify text! ***";
    print STDERR "\n\n";
    help();
  }
  main_cmdline();
} else {
  gui();
  main Gtk2;
}
exit;

# Graphical frontend
sub gui() {
  Gtk2->init;

  $xml_root = Gtk2::GladeXML->new( $glade_dir."/joellisme.glade" );
  $xml_root->signal_autoconnect_from_package ('main');

  my $main_win = $xml_root->get_widget ("gui");
#  $main_win->set_title ("$NAME $config::version");
  $main_win->set_title ("$NAME $VERSION");

  # Create Gtk2::TextTags
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
  $input->create_tag( "ignore",
                      "foreground" => "orange",
                      "style" => "italic" );
  $input->create_tag( "suggest",
#                      "weight" => PANGO_WEIGHT_BOLD,
                      "foreground" => "blue",
                      "style" => "italic" );
  $input->create_tag( "highlight",
                      "weight" => PANGO_WEIGHT_BOLD,
                      "background" => "yellow" );
}

# Menu actions
##############
# File::New
sub menuNew() {
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
  $input->delete( $input->get_start_iter(), $input->get_end_iter() );
}

# File::Open
sub menuOpen() {
  my $fileChooser = $xml_root->get_widget("diagFileChooser");
  my $filter = Gtk2::FileFilter->new();
  $filter->set_name( 'text' );
  $filter->add_mime_type( 'text/plain' );
  $fileChooser->add_filter( $filter );
  $fileChooser->set( 'action' => 'open' );
  msgDebug( "Mode: " . $fileChooser->get( 'action' ) );
  $fileChooser->show();

  my $run = $fileChooser->run();
  if ('ok' eq $run ) {
      my $filename = $fileChooser->get_filename();
      msgDebug( "Filename: " . $filename );
      open( FC, $filename );
      my @text = <FC>;
      close(FC);

      my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
      $input->set_text( join( "", @text ) );
  }

  $fileChooser->hide();
}

# File::Save
sub menuSave() {
  my $fileChooser = $xml_root->get_widget("diagFileChooser");
  my $filter = Gtk2::FileFilter->new();
  $filter->set_name( 'text' );
  $filter->add_mime_type( 'text/plain' );
  $fileChooser->add_filter( $filter );
  $fileChooser->set( 'action' => 'save' );
  msgDebug( "Mode: " . $fileChooser->get( 'action' ) );
  $fileChooser->show();

  my $oldfilename = $fileChooser->get_filename();

  # Filename already defined
  if ( defined($oldfilename) ) {
      msgDebug( "Old filename: " . $oldfilename );
      # Open file in write mode
      open( FC, ">$oldfilename" );
      my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
      # Get text and ignore hidden characters
      print FC $input->get_text( $input->get_start_iter(),
                                 $input->get_end_iter(),
                                 FALSE );
      close(FC);
  } else {
      msgDebug( "No old filename." );
      # Ask for a filename
      my $run = $fileChooser->run();
      if ('ok' eq $run ) {
          my $filename = $fileChooser->get_filename();
          msgDebug( "Filename: " . $filename );

          # Open file in write mode
          open( FC, ">$filename" );
          my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
          # Get text and ignore hidden characters
          print FC $input->get_text( $input->get_start_iter(),
                                     $input->get_end_iter(),
                                     FALSE );
          close(FC);
      }
  }

  $fileChooser->hide();
}

# File::SaveAs
sub menuSaveAs() {
  my $fileChooser = $xml_root->get_widget("diagFileChooser");
  $fileChooser->unselect_all();
  $fileChooser->set_current_name("");
  menuSave();
}

# File::Quit
sub menuQuit() {
  Gtk2->main_quit;
}

# Edit::Couper
sub menuCutText() {
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
  my $clipboard = Gtk2::Clipboard->get( Gtk2::Gdk->SELECTION_PRIMARY );
  $input->cut_clipboard( $clipboard, TRUE );
}

# Edit::Copier
sub menuCopyText() {
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
  my $clipboard = Gtk2::Clipboard->get( Gtk2::Gdk->SELECTION_PRIMARY );
  $input->copy_clipboard( $clipboard );
}

# Edit::Coller
sub menuPasteText() {
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();
  my $clipboard = Gtk2::Clipboard->get( Gtk2::Gdk->SELECTION_PRIMARY );
  # clipboard : the GtkClipboard to paste from
  # override_location : location to insert pasted text, or undef for at the cursor
  # default_editable : 	whether the buffer is editable by default
  $input->paste_clipboard( $clipboard, undef, TRUE );
}

# Edit::Delete
sub menuDeleteText() {
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();

  # interactive : whether the deletion is caused by user interaction
  # default_editable : whether the buffer is editable by default
  $input->delete_selection( TRUE, TRUE );
}

# Help::About
sub menuAbout() {
  my $credit = __"Pascal Giard <evilynux\@gmail.com>\n\n".
  		"Mention spciale  ma douce moiti\n".
  		"Jolle Comtois sans qui cette application n'aurais\n".
  		"pas vu le jour!";

  Gtk2->show_about_dialog(
                          $xml_root->get_widget('gui'),
                          'name' => $NAME,
                          'version' => $VERSION,
                          'logo-icon-name' => 'gtk-find-and-replace',
                          'authors' => $credit,
                          'copyright' => '(c) Pascal Giard, 2006',
                          'comments' => __"Tool aiming toward anglicism eradication.",
                          'license' => __"Available under the terms of the GNU GPL.\n".
                          "See http://www.gnu.org/licenses/gpl.txt",
                         );
}

sub btnSearchClicked() {
  my $textview = $xml_root->get_widget("txtViewInput");
  my $input = $xml_root->get_widget("txtViewInput")->get_buffer();

  # Update future corrected text
  $text = $input->get_text( $input->get_start_iter,
                            $input->get_end_iter, TRUE);

  my %anglicism = process($text);

  msgDebug( "Found " . $anglicism{'nbr'} . " distinct anglicism." );

  # Show anglicism
  for( my $i = 1; $i <= $anglicism{'nbr'}; $i++ ) {
      my %vals = %{$anglicism{'words'}{$i}};

      msgDebug( "Anglicism: " . $vals{'word'} );
      msgDebug( "  Description: " . $vals{'description'} );
      msgDebug( "  Suggestion: " . $vals{'suggest'} );

      # Find anglicism in the textview
      my $iter = $textview->get_iter_at_location(0,0);
      # Anglicism coordinates
      my( $start, $end ) = $iter->forward_search( $vals{'word'}, 'text-only' );
      # Exit condition flag
      my $done = FALSE;

      # Loop while we find copies of that anglicism in the text
      while( !$done ) {
          # No other matches found, move on to next anglicism
          if( !defined($start) || !defined($end) ) {
              $done = TRUE;
              last;
          }
          # Update iter position to current anglicism
          $iter->set_offset( $end->get_offset() );

          msgDebug("Found match at " . $start->get_offset() );
          # Highligh and insert "suggest" if not already done
          unless( $start->has_tag($input->get_tag_table()->lookup("ignore")) ) {
              # Highlight current anglicism and scroll to it
              $input->apply_tag_by_name( "highlight", $start, $end );
              $textview->scroll_to_iter( $start, 0.0, FALSE, 0, 0 );
              # Ask the user what to do about that anglicism
              my $replace = showSuggestion( %vals );
              # Remove highlight
              $input->remove_tag_by_name( "highlight", $start, $end );

              # Apply changes if user doesn't choose to ignore
              unless( $replace eq "" ) {
                  msgDebug("Replace " . $vals{'word'} . " with " . $replace );
                  $input->delete( $start, $end );
                  # Backup old position
                  my $old_offset = $end->get_offset();
                  $input->insert_with_tags_by_name( $end, $replace, "suggest" );
                  # Update iter since textbuffer changed
                  $iter = $textview->get_iter_at_location(0,0);
                  # Move to previous position
                  $iter->forward_chars( $old_offset - length($vals{'word'})
                                        + length($replace) );
              } else {
                  # The user choose to ignore the anglicism
                  msgDebug("Ignore match at " . $start->get_offset() );
                  $input->apply_tag_by_name( "ignore", $start, $end );
              }
          } else {
              # The user already choose to ignore that one on last
              # btnSearchClicked event.
              msgDebug("Ignore match at " . $start->get_offset() );
          }
          # Find next match in text
          ( $start, $end ) = $iter->forward_search( $vals{'word'}, 'text-only' );
      }
  }
}

# Catch the string received as a parameter at the cmdline
sub parseText($) {
  my ($string) = @_;

  if ( defined($string) ) {
    $text = $string;
    msgDebug("From user input: " . $string);
  }
}

# Debug messages
sub msgDebug($) {
  return unless defined($opt{'debug'});
  my ( $string ) = @_;
  print STDERR "Debug: $string\n";
}

# Debug messages
sub msgError($) {
  my ( $string ) = @_;
  print STDERR "ERROR: $string\n";
  exit(1);
}

# Find anglicism
sub process($) {
  my ($string) = @_;
  my %anglicism;

  msgDebug( "Text: " . $text );
  $anglicism{'nbr'} = 0;

  open( FH, "dic.txt" );
  my @dicwords = <FH>;
  close( FH );
  foreach my $word ( @dicwords ) {
      $word =~ s/:(.+)//g;
      my $suggest = $1;
      my $description = __"Not available.";
      if( $suggest =~ /:(.+)$/ ) {
          $suggest =~ s/:(.+)$//g;
          $description = $1;
      }
      chomp($word);
      msgDebug( "Looking for: " . $word );
      if( $text =~ /(\Q$word\E)/g ) {
          $anglicism{'nbr'}++;
          $anglicism{'words'}{$anglicism{'nbr'}}{'word'} = $word;
          $anglicism{'words'}{$anglicism{'nbr'}}{'suggest'} = $suggest;
          $anglicism{'words'}{$anglicism{'nbr'}}{'description'} = $description;
      }
  }

  return %anglicism;
}

# Show anglicism (expression), description and suggestion(s) to user
# Param:
#  %vals: contains keys word, description and suggest
sub showSuggestion(%) {
    my ( %vals ) = @_;
    my $dialog = $xml_root->get_widget("diagReplace");
    my $suggest = $xml_root->get_widget("cboxSuggestion");
    my $suggestion = "";
    $dialog->show();

    # Update anglicism expression and description
    $xml_root->get_widget("lblExpression")->set_text( $vals{'word'} );
    $xml_root->get_widget("lblDescription")->set_text( $vals{'description'} );
    # Split suggestions into a list
    my @suggests = split( ",", $vals{'suggest'} );

    # Add the suggestions to the combobox
    foreach( @suggests  ) {
        $suggest->append_text( $_ );
    }
    # Defaulting to first suggestion
    $suggest->set_active(0);

    # I have to do this!?
    $xml_root->get_widget("diagReplace")->set_title(__"Detected anglicism...");
#    $xml_root->get_widget("btnIgnore")->set_label(__"Ignore");
    $xml_root->get_widget("btnReplace")->set_label(__"Replace");

    # Ask the user what to do next
    my $run = $dialog->run();
    if ('ok' eq $run ) {
        # Replace anglicism
        msgDebug( "Selected: " . $suggest->get_active_text );
        $suggestion = $suggest->get_active_text;
    } else {
        # Ignore anglicism
        msgDebug( "Ignore " . $vals{'word'} );
    }

    $dialog->hide;

    # Clear combobox content
    for( my $i = scalar(@suggests) - 1; $i >= 0 ; $i-- ) {
        $suggest->remove_text( $i );
    }

    return $suggestion;
}

sub test($$) {
  my ( $string, $expected ) = @_;
  $string = process($string);
  print "Got:\t\t\"$string\"\nExpected:\t\"$expected\"\n"
    if( $string !~ /^\Q${expected}\E$/ );
}

# Some use cases in order to make sure nothing gets broken
# FIXME: Add tests for ~ ~= ~~ >= <=
sub selfTest() {
  return unless defined($opt{'debug'});

#  test("alpha", "%alpha ");
}

# Basic help message
sub help() {
  print "Usage: " . $0 . " [options] \"text\"
Note: All options that may be abbreviated to uniqueness
      can use a single dash.

Required argument:

 \"text\"          Your text.

Options:
   --nogui                 Command line mode
   --debug                 Debug mode
   --version               Shows version number
   --help                  Shows this very help message\n\n";

  exit;
}

# Prints version then exits
sub version() {
#   print "$NAME $config::version\n";
  print "$NAME $VERSION\n";
  exit;
}

# Core of the application
sub main_cmdline() {
  print process($text) . "\n";
}
